// $Header: /cvs/gnusto/src/xpcom/engine/gnusto-engine.js,v 1.79 2004/01/03 03:46:24 marnanel Exp $
// Copyright (c) 2004 Thomas Thurman
// thomas@thurman.org.uk
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of version 2 of the GNU General Public License
// as published by the Free Software Foundation.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have be able to view the GNU General Public License at 
// http://www.gnu.org/copyleft/gpl.html ; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.


// not sure how many of these we'll need-- factor these out eventually
const nsIWindowMediator    = Components.interfaces.nsIWindowMediator;
const nsIAppShellService   = Components.interfaces.nsIAppShellService;
const nsICmdLineHandler    = Components.interfaces.nsICmdLineHandler;
const nsICategoryManager   = Components.interfaces.nsICategoryManager;
const nsIContentHandler    = Components.interfaces.nsIContentHandler;
const nsIProtocolHandler   = Components.interfaces.nsIProtocolHandler;
const nsIURI               = Components.interfaces.nsIURI;
const nsIURL               = Components.interfaces.nsIURL;
const nsIStringInputStream = Components.interfaces.nsIStringInputStream;
const nsIChannel           = Components.interfaces.nsIChannel;
const nsIRequest           = Components.interfaces.nsIRequest;
const nsIProgressEventSink = Components.interfaces.nsIProgressEventSink;
const nsISupports          = Components.interfaces.nsISupports;

const well_known_hosts = {
  'groggs': 'rgtp-serv.groggs.group.cam.ac.uk',
  // Let me know of any others that should be here -- Marnanel
};

////////////////////////////////////////////////////////////////
//
// Takes out entities.
//

function html_prepare(text) {
  return String(text).
    replace('&','&amp;','g').
    replace('<','&lt;','g').
    replace('>','&gt;','g');
}

////////////////////////////////////////////////////////////////
//

////////////////////////////////////////////////////////////////
//
// Returns a pretty string representing the date |t|. |t| is
// expressed in seconds since midnight, 1/1/1970.

var months = ['Jan', 'Feb', 'Mar',
	      'Apr', 'May', 'June',
	      'July', 'Aug', 'Sept',
	      'Oct', 'Nov', 'Dec'];

function prettyDate(t) {
  var clock = new Date(t*1000);
  var today = new Date();

  if (clock.getFullYear()==today.getFullYear()) {
    // it was this year

    var hours = clock.getHours();
    var ampm;
    if (hours > 11) { ampm = 'pm' } else { ampm = 'am' }
    if (hours > 12) { hours = hours - 12 }

    var mins = String(clock.getMinutes());
    if (mins.length==1) { mins = '0' + mins; }

    return hours + ':' + mins + ampm + ', ' +
      clock.getDate() + ' ' + months[clock.getMonth()];

  } else {

    // a previous year

    return clock.getDay()+' '+months[clock.getMonth()] + ' ' + clock.getFullYear();
  }
}

////////////////////////////////////////////////////////////////
//
// Class and contract IDs
//

const RGTP_PROTOCOL_HANDLER_CID   = Components.ID("{f483b8ea-b0a8-47dc-8fe2-8c33d2acdb32}");
const RGTP_PROTOCOL_HANDLER_CTRID = "@mozilla.org/network/protocol;1?name=rgtp";

////////////////////////////////////////////////////////////////
//
// The channel-- the source of RGTP data.
//

function RGTPChannel (uri)
{
  this.URI = uri;
  this.originalURI = uri;
  this._isPending = true; // why?
  this.stringStream = Components.classes["@mozilla.org/io/string-input-stream;1"].
    createInstance(nsIStringInputStream);
}

RGTPChannel.prototype.QueryInterface =
function jsdch_qi (iid)
{
  if (!iid.equals(nsIChannel) /* && !iid.equals(nsIRequest) (do we?) */ &&
      !iid.equals(nsISupports)) {
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }

  return this;
}

/* nsIChannel */
RGTPChannel.prototype.loadAttributes = null;
RGTPChannel.prototype.contentType = "text/html";
RGTPChannel.prototype.contentLength = -1;
RGTPChannel.prototype.owner = null;
RGTPChannel.prototype.loadGroup = null;
RGTPChannel.prototype.notificationCallbacks = null;
RGTPChannel.prototype.securityInfo = null;

RGTPChannel.prototype.open =
function rgtpch_open()
{
  // We don't have to provide this (should we, anyway?)
  throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
}

// The design of the following few functions is
// based on that in Venkman.

RGTPChannel.prototype.asyncOpen =
function rgtpch_aopen (streamListener, context)
{
  this.streamListener = streamListener;
  this.context = context;

  var path = this.URI.path;

  try {
    if (path=='/index') {
      this.produce_index();
    } else {
      this.produce_item(path.substring(1));
    }
  } catch(ex) {
    dump('EXCEPTION: ');
    dump(ex);
    dump('\n');
  }
}

var clients = {};

RGTPChannel.prototype.ensure_client =
function rgtpch_ensure_client () {

  this.normalise_host_and_port();

  var host = this.URI.host;
  var port = this.URI.port;
  var key = String(host)+'_'+String(port);

  // FIXME: also if the current one is closed

  if (typeof clients[key]=="undefined") {

    var client = Components.classes["@marnanel.org/archangel/client;1"].
      createInstance(Components.interfaces.nsIReverseGossip);

    client.connect(host, port, '', '');
    // FIXME: and check it

    clients[key] = client;
  }

  this.client = clients[key];
}
RGTPChannel.prototype.produce_index =
function rgtpch_produce_index () {
  // somehow: alert('We have to download the WHOLE index. This will take a while.');

  try {

    var fields = {};
    var URLprefix = this.URI.prePath + '/';

    this.ensure_client();
    this.ensure_style();

    fields.SERVER = this.URI.prePath;

    var sorted = [];
    var odd = true;

    for (var i=0; i<this.client.itemCount; i++) {
      this.client.loadItemByNumber(i);

      var key = this.client.latestSeqKnown.toString(16);
      while (key.length<8) {
	key = '0'+key;
      }
      sorted.push(key + ' ' + this.client.itemID);
    }

    sorted.sort();
    sorted.reverse();

    fields.ENTRIES = '';

    for (k in sorted) {

      var entryfields = {};

      this.client.loadItemByID(sorted[k].substring(9));

      var is_unread = (this.client.latestSeqKnown != this.client.latestSeqRead);

      fields.ENTRIES += this.boilerplate('index.entry', {
	'ITEMURL': URLprefix+this.client.itemID,
	'SUBJECT': html_prepare(this.client.subject),
	'REPLIES': this.client.entryCount-1,
	'LATEST_AUTHOR': this.client.latestBy,
	'ODDEVEN': odd? 'odd' : 'even',
	'READ': is_unread? 'unread':'read',
      });

      odd = !odd;
    }

    this.respond(this.boilerplate('header', fields)+
		 this.boilerplate('index.main', fields));

  } catch(ex) {
    dump('EXCEPTION: ');
    dump(ex);
    dump('\n');
  }
}

RGTPChannel.prototype.normalise_host_and_port =
function rgtpch_normalise_host_and_port() {
    if (well_known_hosts[this.URI.host]) {
      this.URI.host = well_known_hosts[this.URI.host];
    }
    
    if (this.URI.port==-1) {
      // use the IANA default
      this.URI.port = 1431;
    }
}

RGTPChannel.prototype.produce_item =
function rgtpch_produce_item (itemid) {

  try {

    var fields = {};

    this.ensure_client();
    this.ensure_style();

    var URLprefix = this.URI.prePath + '/';

    this.client.loadItemByID(this.URI.path.substring(1));
    this.client.markItemAsRead();    

    var odd = true;

    fields.ENTRIES = '';

    for (var i=0; i<this.client.entryCount; i++) {

      var entryfields = {
      'BODY': '',
      'USERID': '',
      };

      this.client.loadEntry(i);

      var grogname = '';
      var header = true;
      var body = html_prepare(this.client.body).split('\n');

      for (var line in body) {

	var thisline = body[line];

	if (header) {
	  if (thisline=='') {
	    header = false;
	  } else if (entryfields.USERID=='') {
	    var frompos = thisline.indexOf('from');
	    if (entryfields.USERID!=-1) {
	      entryfields.USERID = thisline.substring(frompos+5, thisline.length-23);

	      if (entryfields.USERID.substring(entryfields.USERID.length-1)+']') {
		// Grogname is in with the userid (short form)
		var openbracketpos = entryfields.USERID.lastIndexOf(' (');
		if (openbracketpos!=-1) {
		  var temp = entryfields.USERID.substring(openbracketpos+2,
							  entryfields.USERID.length-1);
		  grogname = entryfields.USERID.substring(0, openbracketpos);
		  entryfields.USERID = temp;
		}
	      }
	    }
	  } else if (thisline.substring(0,5)=='From ') {
	    // Grogname is separate from the userid (long form)
	    grogname = thisline.substring(5);
	  }
	} else {
	  entryfields.BODY += thisline +'\n';
	}
      }

      if (entryfields.BODY!='') {
	
 	entryfields.BODY = entryfields.BODY.
	  replace('\n\n','</p><p>','g').
	  replace('  ',' &nbsp;','g');

	entryfields.ODDEVEN = (odd?"odd":"even");

	if (grogname) {
	  entryfields.GROGNAME = this.boilerplate('item.grogname',
	      {'NAME': grogname});
	} else {
	  entryfields.GROGNAME = '';
	}

	entryfields.PRETTYDATE = prettyDate(this.client.datestamp);

	fields.ENTRIES += this.boilerplate('item.entry', entryfields);

	odd = !odd;
      }

    }    

    fields.SUBJECT = html_prepare(this.client.subject);
    fields.INDEX = URLprefix+'index';

    var parent = this.client.parent;
    var child = this.client.child;

    if (parent) {
      this.client.loadItemByID(parent);
      fields.CONTINUED_FROM = this.boilerplate('item.continued_from',
	  {'PARENT_URL': URLprefix+parent,
	   'PARENT_SUBJECT': this.client.subject,
	  });
    } else {
      fields.CONTINUED_FROM = '';
    }
    
    if (child) {
      this.client.loadItemByID(child);
      fields.CONTINUED_IN = this.boilerplate('item.continued_in',
	  {'CHILD_URL': URLprefix+child,
	   'CHILD_SUBJECT': this.client.subject,
	  });
    } else {
      fields.CONTINUED_IN = '';
    }

    this.client.markItemAsRead();
    this.respond(this.boilerplate('header', fields)+
		 this.boilerplate('item.main', fields));

  } catch(ex) {
    dump('EXCEPTION: ');
    dump(ex);
    dump('\n');
  }
}

RGTPChannel.prototype.style = undefined;

RGTPChannel.prototype.ensure_style =
function rgtpch_ensure_style() {
  if (typeof this.style == 'object') return;

  this.style = {};

  var bundle = Components.Constructor('@mozilla.org/intl/stringbundle;1',
				      'nsIStringBundleService')().
    createBundle('chrome://archangel/skin/style.ini').
    getSimpleEnumeration();

  while(bundle.hasMoreElements()) {
    var prop = bundle.getNext().QueryInterface(Components.interfaces.nsIPropertyElement);
    this.style[prop.key] = prop.value;
  }
}

RGTPChannel.prototype.boilerplate =
function rgtpch_boilerplate(key, fields) {
  var result = this.style[key];

  for (field in fields) {
    result = result.replace('${'+field+'}', fields[field], 'g');
  }

  return result;
}

RGTPChannel.prototype.respond =
function rgtpch_respond (str)
{
    this.streamListener.onStartRequest (this, this.context);

    var len = str.length;
    this.stringStream.setData (str, len);
    this.streamListener.onDataAvailable (this, this.context,
                                         this.stringStream, 0, len);
    this.streamListener.onStopRequest (this, this.context,
                                       Components.results.NS_OK);

    /* ???
       if (this.loadGroup)
        this.loadGroup.removeRequest (this, null, Components.results.NS_OK);
    */
    this._isPending = false;    
}

/* nsIRequest -- do we do this?
RGTPChannel.prototype.isPending =
function rgtpch_ispending ()
{
    return this._isPending;
}

RGTPChannel.prototype.status = Components.results.NS_OK;

RGTPChannel.prototype.cancel =
function rgtpch_cancel (status)
{
    if (this._isPending)
    {
        this._isPending = false;
        this.streamListener.onStopRequest (this, this.context, status);
        if (this.loadGroup)
        {
            try
            {
                this.loadGroup.removeRequest (this, null, status);
            }
            catch (ex)
            {
                debug ("we're not in the load group?\n");
            }
        }
    }
    
    this.status = status;
}

RGTPChannel.prototype.suspend =
RGTPChannel.prototype.resume =
function rgtpch_notimpl ()
{
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
} */


////////////////////////////////////////////////////////////////
//
// The protocol handler.
//

function RGTPProtocolHandler() {
}


RGTPProtocolHandler.prototype.scheme = "rgtp";
//RGTPProtocolHandler.prototype.URI = ... "rgtp";

RGTPProtocolHandler.prototype.defaultPort = 1431; // IANA default
// What kind of protocol are we?
// Well, we support authority; we allow relative URIs;
// we can't be proxied by anything on earth. So that's the default.
RGTPProtocolHandler.prototype.protocolFlags = nsIProtocolHandler.URI_STD;

// No, we don't expect people to play funny games with this protocol.
RGTPProtocolHandler.prototype.allowPort =
function jsdph_allowport (aPort, aScheme) { return false; }

// And now the meat:

// Turns a string into an nsURI, possibly relative to
// another nsURI.
RGTPProtocolHandler.prototype.newURI =
function jsdph_newuri (spec, charset, baseURI)
{

  // FIXME: Need to deal with baseURI.
  var uri = Components.classes["@mozilla.org/network/standard-url;1"].
    createInstance(Components.interfaces.nsIURI);

  uri.spec = spec;

  return uri;
}

// Turns an nsURI into an nsIChannel
RGTPProtocolHandler.prototype.newChannel =
function jsdph_newchannel (uri)
{
  return new RGTPChannel (uri);
}


////////////////////////////////////////////////////////////////
//
// The factories:
//
// RGTP protocol handler factory.
//

var RGTPProtocolHandlerFactory = new Object();

RGTPProtocolHandlerFactory.createInstance =
function rgtp_phf_create(outer, iid)
{
  if (outer != null) {
    dump("Don't squish us!\n");
    throw Components.results.NS_ERROR_NO_AGGREGATION;
  }

  if (!iid.equals(nsIProtocolHandler) && !iid.equals(nsISupports)) {
    dump("Ugh, weird interface.\n");
    throw Components.results.NS_ERROR_INVALID_ARG;
  }
  
  return new RGTPProtocolHandler();
}

////////////////////////////////////////////////////////////////
//
// RGTP content handler factory.
//

var RGTPContentHandlerFactory = new Object();

RGTPContentHandlerFactory.createInstance =
function rgtp_chf_create(outer, iid)
{
  // what's this for?
  if (outer != null)
    throw Components.results.NS_ERROR_NO_AGGREGATION;

  if (!iid.equals(nsIContentHandler) && !iid.equals(nsISupports))
    throw Components.results.NS_ERROR_INVALID_ARG;

  return new RGTPContentHandler();
}


////////////////////////////////////////////////////////////////
//
// The Module
//

var Module = new Object();

Module.registerSelf =
function mod_regself(compMgr, fileSpec, location, type) {
  reg = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  reg.registerFactoryLocation(RGTP_PROTOCOL_HANDLER_CID,
			      "RGTP protocol handler",
			      RGTP_PROTOCOL_HANDLER_CTRID,
			      fileSpec,
			      location,
			      type);
}

Module.getClassObject =
function mod_getclassobj(compMgr, cid, iid) {
  if (cid.equals(RGTP_PROTOCOL_HANDLER_CID)) {
    return RGTPProtocolHandlerFactory;
  }

  // okay, so something's weird. give up.
  
  if (!iid.equals(Components.interfaces.nsIFactory))
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  
  throw Components.results.NS_ERROR_NO_INTERFACE;
  
}

Module.canUnload =
function mod_canunload(compMgr) {
  return true;
}

////////////////////////////////////////////////////////////////
//
// NSGetModule
//
// It all starts here!

function NSGetModule(compMgr, fileSpec) {
  return Module;
}
